#!/usr/bin/env python

from datetime import datetime
import logging

from bardolph.lib import injection
from bardolph.controller import i_controller


class Light:
    def __init__(self, name, group, location, color, multizone):
        self._name = name
        self._group = group
        self._location = location
        self._multizone = multizone
        self._power = 12345
        self._color = color if color is not None else [0, 0, 0, 0]

    def __repr__(self):
        rep = '_name: "{}", _group: "{}", _location: "{}", _power: {}, '.format(
            self._name, self._group, self._location, self._power)
        rep += '_color: {}'.format(self._color)
        return rep

    def get_color(self):
        logging.info('Get _color "{}" {}'.format(self._name, self._color))
        return self._color

    def set_color(self, color, duration=0, _=False):
        self._color = color
        logging.info('Color "{}" {}, {}'.format(self._name, color, duration))

    def set_zone_color(self, start_index, end_index, color, duration, _=False):
        logging.info('Multizone color "{}" {}, {}, {}, {}'.format(
            self._name, start_index, end_index, color, duration))

    def supports_multizone(self):
        return self._multizone
    
    def set_return_color(self, color):
        self._color = color

    def set_power(self, power, duration):
        self._power = power
        logging.info('Power "{}", {}, {}'.format(self._name, power, duration))

    def get_power(self):
        return self._power

    def get_color_zones(self, start=0, end=7):
        return [[self._color]] * (end - start)

    def set_return_power(self, power):
        self._power = power

    def get_label(self):
        return self._name

    def get_location(self):
        return self._location

    def get_group(self):
        return self._group


class LightSet:
    def __init__(self):
        self.clear_lights()
        self.discover()

    def clear_lights(self):
        self._lights = {}
        self._groups = {}
        self._locations = {}

    def discover(self):
        if len(self._lights) == 0:
            self.add_light('Table', 'Furniture', 'Home', [1, 2, 3, 4])
            self.add_light('Top', 'Pole', 'Home', [10, 20, 30, 40])
            self.add_light('Middle', 'Pole', 'Home', [100, 200, 300, 400])
            self.add_light(
                'Bottom', 'Pole', 'Home', [1000, 2000, 3000, 4000])
            self.add_light(
                'Chair', 'Furniture', 'Home', [10000, 20000, 30000, 40000])
            self.add_light('Strip', 'Furniture', 'Home', [4, 3, 2, 1], True)

    def add_light(self, name, group, location, color=None, multizone=False):
        new_light = Light(name, group, location, color, multizone)
        self._lights[name] = new_light
        if group is not None:
            if group in self._groups:
                self._groups[group].append(new_light)
            else:
                self._groups[group] = [new_light]
        if location is not None:
            if location in self._locations:
                self._locations[location].append(new_light)
            else:
                self._locations[location] = [new_light]

    @property
    def light_names(self):
        return self._lights.keys()

    @property
    def lights(self):
        return list(self._lights.values())

    @property
    def group_names(self):
        return self._groups.keys()

    @property
    def location_names(self):
        return self._locations.keys()
    
    @property
    def count(self):
        return len(self._lights)

    @property
    def last_discover(self):
        return datetime.now()

    @property
    def successful_discovers(self):
        return 100

    @property
    def failed_discovers(self):
        return 10

    def get_light(self, name):
        if not name in self._lights:
            logging.error("Light >\"{}\"< not in fake _lights".format(name))
            return None
        return self._lights[name]

    def get_group(self, name):
        return self._groups.get(name, {})

    def get_location(self, name):
        return self._locations.get(name, {})

    def set_color(self, color, duration):
        logging.info("Color (all) {}, {}".format(color, duration))
        
    def get_color(self):
        return [1, 2, 3, 4]

    def set_power(self, power_level, duration):
        logging.info("Power (all) {} {}".format(power_level, duration))

    def get_power(self):
        return True

def configure():
    injection.bind_instance(LightSet()).to(i_controller.LightSet)



class UnboundException(Exception): pass


class Injection:
    providers = {}


class Binder:
    """ Parameter implementation is a class to be used as the factory. """
    def __init__(self, implementation):
        self._implementation = implementation

    def to(self, interface):
        Injection.providers[interface] = self._implementation


class ObjectBinder:
    """ Singleton. Parameter implementor is always returned. """
    def __init__(self, implementor):
        self._implementor = implementor

    def to(self, interface):
        Injection.providers[interface] = lambda: self._implementor


def inject(*interfaces):
    def fn_wrapper(fn):
        def param_wrapper(*args):
            injected_args = []
            for interface in interfaces:
                if interface in Injection.providers:
                    obj = Injection.providers[interface]()
                else:
                    obj = interface()
                injected_args.append(obj)
            args = args + tuple(injected_args)
            return fn(*args)
        return param_wrapper
    return fn_wrapper


def provide(interface):
    if interface not in Injection.providers:
        raise UnboundException("interface {}".format(interface))

    creator = Injection.providers[interface]
    return creator()


def configure(): Injection.providers = {}

def bind(implementation): return Binder(implementation)

def bind_instance(implementor): return ObjectBinder(implementor)



from enum import Enum

from bardolph.lib.auto_repl import auto


class OpCode(Enum):
    COLOR = auto()
    END = auto()
    GET_COLOR = auto()
    NOP = auto()
    PAUSE = auto()
    POWER = auto()
    SET_REG = auto()
    STOP = auto()
    TIME_PATTERN = auto()
    TIME_WAIT = auto()
    ZONE = auto()


class Operand(Enum):
    ALL = auto()
    LIGHT = auto()
    GROUP = auto()
    LOCATION = auto()
    MZ_LIGHT = auto()


class SetOp(Enum):
    """ Used with TimePattern """
    INIT = auto()
    UNION = auto()
    

class Register(Enum):
    HUE = auto()
    SATURATION = auto()
    BRIGHTNESS = auto()
    KELVIN = auto()
    DURATION = auto()
    POWER = auto()
    NAME = auto()
    OPERAND = auto()
    TIME = auto()
    ZONES = auto()


class Instruction:
    def __init__(self, op_code, param0=None, param1=None):
        self._op_code = op_code
        self._param0 = param0
        self._param1 = param1

    def __repr__(self):
        if self._op_code == OpCode.TIME_PATTERN:
            return 'Instruction({}, {}, {})'.format(
                OpCode.TIME_PATTERN, self.param0, repr(self.param1))
        if self._param1 is None:
            if self._param0 is None:
                return 'Instruction({}, None, None)'.format(self._op_code)
            return 'Instruction({}, {}, None)'.format(
                self._op_code,
                Instruction.quote_if_string(self._param0))
        return 'Instruction({}, {}, {})'.format(
            self._op_code,
            Instruction.quote_if_string(self._param0),
            Instruction.quote_if_string(self._param1))
        
    def __eq__(self, other):
        if not isinstance(other, type(self)):
            raise TypeError
        return (self._op_code == other._op_code
                and self._param0 == other._param0
                and self._param1 == other._param1)

    @property
    def op_code(self):
        return self._op_code
    
    @property
    def param0(self):
        return self._param0
    
    @property
    def param1(self):
        return self._param1
    
    def as_list_text(self):
        if self._op_code not in (OpCode.SET_REG, OpCode.TIME_PATTERN):
            return '{}'.format(self._op_code)
        return '{}, {}, {}'.format(
            self._op_code,
            Instruction.quote_if_string(self._param0),
            Instruction.quote_if_string(self._param1))

    @classmethod
    def quote_if_string(cls, obj):
        return ('"{}"' if isinstance(obj, str) else '{}').format(obj)



from enum import Enum

from bardolph.lib.auto_repl import auto

class TokenTypes(Enum):
    ALL = auto()
    AND = auto()
    AT = auto()
    DEFINE = auto()
    EOF = auto()
    GET = auto()
    GROUP = auto()
    LITERAL = auto()
    LOCATION = auto()
    LOGICAL = auto()
    NUMBER = auto()
    OFF = auto()
    ON = auto()
    OR = auto()
    POWER = auto()
    PAUSE = auto()
    RAW = auto()
    REGISTER = auto()
    SET = auto()
    SYNTAX_ERROR = auto()
    TIME_PATTERN = auto()
    UNITS = auto()
    UNKNOWN = auto()
    ZONE = auto()



import configparser

from . import i_lib
from . import injection


class Settings:
    _the_config = {}

    def __init__(self):
        self._config = Settings._the_config
        
    def __contains__(self, name):
        return name in self._config

    def get_value(self, name, default=None):
        return self._config.get(name, default)


class Builder:
    def __init__(self, initial=None):
        self._config = initial.copy() if initial is not None else {}

    def add_overrides(self, overrides):
        self._config.update(overrides)
        return self

    def apply_file(self, file_name):
        if self._config is None:
            self._config = {}
        config = configparser.ConfigParser()
        config.read(file_name)
        for section in config.sections():
            for key in config[section]:
                value = config[section][key]
                self._config[key] = {
                    'True': True,
                    'False': False
                }.get(value, value)
        return self

    def configure(self):
        Settings._the_config = self._config
        injection.bind(Settings).to(i_lib.Settings)
        

def use_base(initial=None):
    return Builder(initial)



import re

from . import i_lib


class TimePattern(i_lib.TimePattern):
    HOURS_24 = set(range(0, 24))
    MINUTES_60 = set(range(0, 60))
    REGEX = re.compile(r'^(\*|\*\d|\d\*|\d\d?):(\d\d|\d\*|\*\d|\*)$')
    
    def __init__(self, hours, minutes):
        self._repr = 'TimePattern("{}", "{}")'.format(hours, minutes)
        self._hour_set = set()
        self._minute_set = set()
        self._init_hour_set(hours) 
        self._init_minute_set(minutes)
    
    def __repr__(self):
        return self._repr
    
    @classmethod
    def from_string(cls, pattern):
        the_match = TimePattern.REGEX.match(pattern)
        if the_match is not None:
            hours, minutes = the_match.groups()
            if TimePattern.patterns_valid(hours, minutes):
                return TimePattern(hours, minutes)
        return None

    @classmethod
    def patterns_valid(cls, hours, minutes):
        return (TimePattern.hours_valid(hours)
                and TimePattern.minutes_valid(minutes))
     
    @classmethod
    def hours_valid(cls, hours):
        if hours == '*':
            return True
        if len(hours) not in (1, 2):
            return False
        if hours[0] == '*':
            return hours[1].isdigit()
        if len(hours) == 2 and hours[1] == '*':
            return hours[0] in '012'
        if not hours.isdecimal():
            return False
        int_hours = int(hours)
        return int_hours >= 0 and int_hours < 25
    
    @classmethod
    def minutes_valid(cls, minutes):
        if minutes == '*':
            return True
        if len(minutes) != 2:
            return False
        if minutes[0] == '*':
            return minutes[1].isdigit()
        if minutes[1] == '*':
            return minutes[0] in '012345'
        if not minutes.isdecimal():
            return False
        int_minutes = int(minutes)
        return int_minutes >= 0 and int_minutes < 60
    
    def union(self, other):
        self._hour_set.update(other._hour_set)
        self._minute_set.update(other._minute_set)

    def match(self, hours, minutes):
        return hours in self._hour_set and minutes in self._minute_set    

    def _init_hour_set(self, pattern):
        if pattern == '*':
            self._hour_set=TimePattern.HOURS_24.copy()
        elif len(pattern) == 1:
            self._hour_set = set([int(pattern)])
        else:
            self._hour_set = set()
            for hour in range(0, 24):
                if TimePattern._number_match(hour, pattern):
                    self._hour_set.add(hour)

    def _init_minute_set(self, pattern):
        if pattern == '*':
            self._minute_set = TimePattern.MINUTES_60.copy()
        else:
            self._minute_set = set()
            for minute in range(0, 59):
                if TimePattern._number_match(minute, pattern):
                    self._minute_set.add(minute)

    @classmethod
    def _number_match(cls, number, pattern):
        formatted = "{:02d}".format(number)
        return (formatted == pattern
            or (pattern[0] in ('*', formatted[0]) 
              and pattern[1] in ('*', formatted[1])))
                    



class Clock:
    def start(self): pass
    def stop(self): pass
    def reset(self): pass
    def pause_for(self, _): pass
    
class Settings: pass

class TimePattern:
    def match(self, hour, minute): pass
    
class LogConfig: pass

def configure(): pass



#!/usr/bin/env python

import argparse
import logging

from bardolph.lib import injection
from bardolph.lib import settings
from bardolph.lib.time_pattern import TimePattern

from bardolph.controller import arg_helper
from bardolph.controller import config_values
from bardolph.controller.instruction import Instruction, OpCode
from bardolph.controller.instruction import Operand, Register, SetOp
from bardolph.controller import light_module
from bardolph.controller import machine

instructions = [
    #instructions

]

current_instruction = 0

def next_instruction():
    global current_instruction
    if current_instruction < len(instructions):
        value = instructions[current_instruction]
        current_instruction += 1
        return value
    return None

def build_instructions():
    program = []
    op_code = next_instruction()
    while op_code != None:
        if op_code in (OpCode.SET_REG, OpCode.TIME_PATTERN):
            param0 = next_instruction()
            param1 = next_instruction()
            program.append(Instruction(op_code, param0, param1))
        else:
            program.append(Instruction(op_code))
        op_code = next_instruction()
    return program

def main():
    ap = argparse.ArgumentParser()
    ap.add_argument(
        '-d', '--debug', help='do debug-level logging', action='store_true')
    ap.add_argument(
        '-f', '--fakes', help='use fake lights', action='store_true')
    arg_helper.add_n_argument(ap)
    args = ap.parse_args()

    injection.configure()
    settings_init = settings.use_base(config_values.functional)    
    overrides = {
        'sleep_time': 0.1
    }
    if args.debug:
        overrides['log_level'] = logging.DEBUG
        overrides['log_to_console'] = True
    if args.fakes:
        overrides['use_fakes'] = True    
    n_arg = arg_helper.get_overrides(args)
    if n_arg is not None and not args.fakes:
        overrides.update(n_arg)

    settings_init.add_overrides(overrides).configure()
    light_module.configure()    
    machine.Machine().run(build_instructions())


if __name__ == '__main__':
    main()



from ..lib import clock
from ..lib import log_config
from ..lib.i_lib import Settings
from ..lib.injection import provide

def configure():
    """ Assumes injection and settings are already initialized. """
    log_config.configure()
    clock.configure()

    settings = provide(Settings)
    if settings.get_value('use_fakes'):
        from ..fakes import fake_light_set
        fake_light_set.configure()
    else:
        from . import light_set
        light_set.configure()



from enum import Enum

from bardolph.lib.auto_repl import auto
from ..controller.instruction import Register


_RAW_RANGE = (0, 65535)


class UnitMode(Enum):
    LOGICAL = auto()
    RAW = auto()


class Units:
    def has_range(self, reg):
        return reg not in (Register.DURATION, Register.TIME)

    def requires_conversion(self, reg):
        return reg not in (
            Register.DURATION, Register.KELVIN, Register.TIME)

    def get_range(self, reg):
        """
        Return the allowable range, in raw units, for a parameter.

        Reg:
            token_type.TokenType designating the register to be set. May also
            be a string containing the name of the Enum.

        Returns:
            A tuple containing (minimum, maximum), or None if the parameter
            does not have a limited range of values.
        """
        reg = Units._string_check(reg)
        return None if reg in (
            Register.DURATION, Register.TIME) else _RAW_RANGE

    def as_raw(self, reg, logical_value):
        """If necessary, converts to integer value that can be passed into the
        light API.

        Args:
            reg: TokenType corresponding to the register being set.
            logical_value: the number to be converted. May also
            be a string containing the name of the Enum.

        Returns:
            If no conversion is done, returns the incoming value untouched.
            Otherwise, an integer that corresponds to the logical value.
        """
        value = logical_value
        reg = Units._string_check(reg)
        if self.requires_conversion(reg):
            if reg == Register.HUE:
                if logical_value in (0.0, 360.0):
                    value = 0
                else:
                    value = round((logical_value % 360.0) / 360.0 * 65535.0)
            elif reg in (Register.BRIGHTNESS, Register.SATURATION):
                if logical_value == 100.0:
                    value = 65535
                else:
                    value = round(logical_value / 100.0 * 65535.0)
        return round(value)

    def as_logical(self, reg, raw_value):
        """If necessary, converts to floating-point logical value that
        typically apears in a script.

        Args:
            reg: TokenType corresponding to the register being set.
            raw_value: the number to be converted. May also
            be a string containing the name of the Enum.

        Returns:
            If no conversion is done, returns the incoming value untouched.
            Otherwise, a float that corresponds to the raw value.
        """
        value = raw_value
        reg = Units._string_check(reg)
        if not self.requires_conversion(reg):
            return raw_value
        if reg == Register.HUE:
            value = float(raw_value) / 65535.0 * 360.0
        elif reg in (Register.BRIGHTNESS, Register.SATURATION):
            if raw_value == 65535:
                value = 100.0
            else:
                value = float(raw_value) / 65535.0 * 100.0
        return value
    
    @classmethod
    def _string_check(cls, reg):
        return Register[reg.upper()] if isinstance(reg, str) else reg



from bardolph.lib import injection, settings
from bardolph.controller import config_values, light_module
from bardolph.lib import job_control
from bardolph.controller.script_runner import ScriptRunner

class LsModule:
    _jobs = None

    @classmethod
    def configure(cls):
        injection.configure()
        settings.using_base(config_values.functional).configure()
        light_module.configure()
        LsModule._jobs = job_control.JobControl()

    @classmethod
    def queue_script(cls, script):
        LsModule._jobs.add_job(ScriptRunner.from_string(script))

    @classmethod
    def request_stop(cls):
        LsModule._jobs.request_stop()


def configure():
    LsModule.configure()


def queue_script(script):
    LsModule.queue_script(script)


def request_stop():
    LsModule.request_stop()



import logging

from lifxlan.errors import WorkflowException

from ..lib.i_lib import Clock, TimePattern
from ..lib.injection import inject, provide

from .get_key import getch
from .i_controller import LightSet
from .instruction import OpCode, Operand, SetOp

class Registers:
    def __init__(self):
        self.hue = 0
        self.saturation = 0
        self.brightness = 0
        self.kelvin = 0
        self.duration = 0
        self.power = False
        self.name = None
        self.operand = None
        self.zones = None
        self.time = 0 # ms.

    def get_color(self):
        return [self.hue, self.saturation, self.brightness, self.kelvin]

    def get_power(self):
        return 65535 if self.power else 0


class Machine:
    def __init__(self):
        self._pc = 0
        self._cue_time = 0
        self._clock = provide(Clock)
        self._variables = {}
        self._program = []
        self._reg = Registers()
        self._enable_pause = True
        self._fn_table = {
            OpCode.COLOR: self._color,
            OpCode.END: self._end,
            OpCode.GET_COLOR: self._get_color,
            OpCode.NOP: self._nop,
            OpCode.PAUSE: self._pause,
            OpCode.POWER: self._power,
            OpCode.SET_REG: self._set_reg,
            OpCode.STOP: self.stop,
            OpCode.TIME_PATTERN: self._time_pattern,
            OpCode.TIME_WAIT: self._time_wait
        }

    def run(self, program):
        self._program = program
        self._pc = 0
        self._cue_time = 0
        self._clock.start()
        while self._pc < len(self._program):
            inst = self._program[self._pc]
            if inst.op_code == OpCode.STOP:
                break
            self._fn_table[inst.op_code]()
            self._pc += 1
        self._clock.stop()

    def stop(self):
        self._pc = len(self._program)

    def color_from_reg(self):
        return [self._reg.hue, self._reg.saturation, self._reg.brightness,
                self._reg.kelvin]

    def color_to_reg(self, color):
        if color is not None:
            reg = self._reg
            reg.hue, reg.saturation, reg.brightness, reg.kelvin = color

    def _color(self): {
        Operand.ALL: self._color_all,
        Operand.LIGHT: self._color_light,
        Operand.GROUP: self._color_group,
        Operand.LOCATION: self._color_location,
        Operand.MZ_LIGHT: self._color_mz_light
    }[self._reg.operand]()

    @inject(LightSet)
    def _color_all(self, light_set):
        light_set.set_color(self._reg.get_color(), self._reg.duration)

    @inject(LightSet)
    def _color_light(self, light_set):
        light = light_set.get_light(self._reg.name)
        if light is None:
            Machine._report_missing(self._reg.name)
        else:
            light.set_color(self._reg.get_color(), self._reg.duration, True)
            
    @inject(LightSet)
    def _color_mz_light(self, light_set):
        light = light_set.get_light(self._reg.name)
        if light is None:
            Machine._report_missing(self._reg.name)
        elif self._zone_check(light):
            start_index, end_index = self._reg.zones
            light.set_zone_color(
                start_index, end_index, 
                self._reg.get_color(), self._reg.duration, True)

    @inject(LightSet)
    def _color_group(self, light_set):
        self._color_multiple(light_set.get_group(self._reg.name))

    @inject(LightSet)
    def _color_location(self, light_set):
        self._color_multiple(light_set.get_location(self._reg.name))

    def _color_multiple(self, lights):
        color = self._reg.get_color()
        for light in lights:
            light.set_color(color, self._reg.duration, True)

    def _power(self): {
        Operand.ALL: self._power_all,
        Operand.LIGHT: self._power_light,
        Operand.GROUP: self._power_group,
        Operand.LOCATION: self._power_location
    }[self._reg.operand]()

    @inject(LightSet)
    def _power_all(self, light_set):
        light_set.set_power(self._reg.get_power(), self._reg.duration)

    @inject(LightSet)
    def _power_light(self, light_set):
        light = light_set.get_light(self._reg.name)
        if light is None:
            Machine._report_missing(self._reg.name)
        else:
            light.set_power(self._reg.get_power(), self._reg.duration, True)

    @inject(LightSet)
    def _power_group(self, light_set):
        self._power_multiple(light_set.get_group(self._reg.name))

    @inject(LightSet)
    def _power_location(self, light_set):
        self._power_multiple(light_set.get_location(self._reg.name))

    def _power_multiple(self, lights):
        power = self._reg.get_power()
        for light in lights:
            light.set_power(power, self._reg.duration)

    @inject(LightSet)
    def _get_color(self, light_set):
        light = light_set.get_light(self._reg.name)
        if light is None:
            Machine._report_missing(self._reg.name)
        else:
            if self._reg.operand == Operand.MZ_LIGHT:
                if self._zone_check(light):
                    zone = self._reg.zones[0]
                    self.color_to_reg(light.get_color_zones(zone, zone + 1))[0]
            else:
                self.color_to_reg(light.get_color())
            
    def _nop(self): pass

    def _pause(self):
        if self._enable_pause:
            print("Press any to continue, q to quit, ! to run.")
            char = getch()
            if char == 'q':
                self.stop()
            else:
                print("Running...")
                if char == '!':
                    self._enable_pause = False

    def _check_wait(self):
        time = self._reg.time
        if isinstance(time, TimePattern):
            self._clock.wait_until(self._reg.time)
        elif time > 0:
            self._clock.pause_for(time / 1000.0)

    def _end(self):
        self.stop()

    def _set_reg(self):
        # param0 is the name of the register, param1 is its value.
        inst = self._program[self._pc]
        setattr(self._reg, inst.param0.name.lower(), inst.param1)

    def _time_pattern(self):
        inst = self._program[self._pc]
        if inst._param0 == SetOp.INIT:
            self._reg.time = inst._param1
        else:
            self._reg.time.union(inst._param1)

    def _time_wait(self):
        self._check_wait()

    def _zone_check(self, light):
        try:
            if not light.supports_multizone():
                logging.warning(
                    'Light "{}" is not multi-zone.'.format(light.get_label()))
                return False
        except WorkflowException:
            name = light.get_label()
            logging.warning(
                'Exception checking capability of "{}"'.format(name))
        return True

    @classmethod
    def _report_missing(cls, name):
        logging.warning("Light \"{}\" not found.".format(name))

    def _power_param(self):
        return 65535 if self._reg.power else 0



#!/usr/bin/env python

import logging
import threading
import time

import lifxlan

from ..lib.color import average_color
from ..lib.injection import bind_instance, inject
from ..lib.i_lib import Settings

from . import i_controller


class LightSet(i_controller.LightSet):
    the_instance = None

    def __init__(self):
        self._light_dict = {}
        self._groups = {}
        self._locations = {}
        self._num_successful_discovers = 0
        self._num_failed_discovers = 0
        self._last_discover = None

    def __len__(self):
        return self.get_count()

    @classmethod
    def configure(cls):
        if LightSet.the_instance is None:
            LightSet.the_instance = LightSet()

    @classmethod
    def get_instance(cls):
        return LightSet.the_instance

    @inject(Settings)
    def discover(self, settings):
        logging.debug("discover()")
        default_num = int(settings.get_value('default_num_lights', 0))
        num_expected = None if default_num == 0 else default_num
        try:
            light_list = lifxlan.LifxLAN(num_expected).get_lights()
            new_dict = {}
            for light in light_list:
                new_dict[light.get_label()] = light
            self._light_dict = new_dict
            self._last_discover = time.localtime()

            self._groups = LightSet._build_set(
                light_list, lifxlan.device.Device.get_group)
            self._locations = LightSet._build_set(
                light_list, lifxlan.device.Device.get_location)
        except lifxlan.errors.WorkflowException as ex:
            logging.error("Error during discovery {}".format(ex))
            return False

        actual = len(self._light_dict)
        if num_expected is not None and actual < num_expected:
            logging.warning(
                "Expected {} devices, found {}".format(num_expected, actual))
            return False
        return True

    @classmethod
    def _build_set(cls, light_list, fn):
        # Produces a dictionary keyed on group or location name, pointing to
        # a list of (lifxlan) light objects.
        sets = {}
        for light in light_list:
            set_name = fn(light)
            if set_name in sets:
                sets[set_name].append(light)
            else:
                sets[set_name] = [light]
        return sets

    @property
    def light_names(self):
        """ list of strings """
        return self._light_dict.keys()

    @property
    def lights(self):
        """ list of Lights. """
        return self._light_dict.values()

    @property
    def group_names(self):
        """ list of strings """
        return self._groups.keys()

    @property
    def location_names(self):
        """ list of strings """
        return self._locations.keys()

    @property
    def count(self):
        return len(self._light_dict)

    @property
    def last_discover(self):
        return self._last_discover

    @property
    def successful_discovers(self):
        return self._num_successful_discovers

    @property
    def failed_discovers(self):
        return self._num_failed_discovers

    def get_light(self, name):
        """ returns an instance of lifxlan.Light, or None if it's not there """
        return self._light_dict.get(name, None)

    def get_group(self, name):
        """ list of Lights """
        return self._groups.get(name, [])

    def get_location(self, name):
        """ list of Lights. """
        return self._locations.get(name, [])

    @inject(Settings)
    def set_color(self, color, duration, settings):
        num_expected = int(settings.get_value('default_num_lights', 0))
        lifx = lifxlan.LifxLAN(num_expected)
        lifx.set_color_all_lights(color, duration, True)
        return True

    def get_color(self):
        """ Returns arithmetic mean of each color setting. """
        return average_color(
            list(self.get_lifxlan().get_color_all_lights().values()))

    def set_power(self, power_level, duration):
        self.get_lifxlan().set_power_all_lights(power_level, duration, True)
        return True

    def get_power(self):
        """ Returns True if at least one bulb is on. """
        for state in self.get_lifxlan().get_power_all_lights().values():
            if state:
                return True
        return False

    @inject(Settings)
    def get_lifxlan(self, settings):
        num_expected = int(settings.get_value('default_num_lights', 0))
        return lifxlan.LifxLAN(num_expected)


def start_light_refresh():
    logging.info("Starting refresh thread.")
    threading.Thread(
        target=light_refresh, name='rediscover', daemon=True).start()


@inject(Settings)
def light_refresh(settings):
    success_sleep_time = float(settings.get_value('refresh_sleep_time'))
    failure_sleep_time = float(settings.get_value('failure_sleep_time'))
    complete_success = False

    while True:
        lights = LightSet.get_instance()
        try:
            complete_success = lights.discover()
            lights._num_successful_discovers += 1
        except lifxlan.errors.WorkflowException as ex:
            logging.warning("Error during discovery {}".format(ex))
            lights._num_failed_discovers += 1

        time.sleep(
            success_sleep_time if complete_success else failure_sleep_time)


@inject(Settings)
def configure(settings):
    LightSet.configure()
    lights = LightSet.get_instance()
    bind_instance(lights).to(i_controller.LightSet)
    lights.discover()

    single = bool(settings.get_value('single_light_discover', False))
    if not single:
        start_light_refresh()



import datetime
import logging

from .i_lib import Settings
from .injection import inject


logging_defaults = {
    'log_date_format': '%D %H:%M:%S',
    'log_file_name': '/tmp/python_log',
    'log_format':
        '%(asctime)s %(filename)s(%(lineno)d) %(funcName)s(): %(message)s',
    'log_level': logging.INFO,
    'log_to_console': False
}

class LogConfig():
    def _extract_level(self):
        level = self.get_value('log_level')
        if isinstance(level, str):
            level = getattr(logging, level.strip(), logging.ERROR)
        return level
    
    @inject(Settings)
    def get_value(self, name, settings):
        return settings.get_value(name, logging_defaults.get(name, None))

    def configure(self):
        if self.get_value('log_to_console'):
            logging.basicConfig(
                level=self._extract_level(),
                format=self.get_value('log_format'),
                datefmt=self.get_value('log_date_format'))
        else:
            logging.basicConfig(
                filename=self.get_value('log_file_name'),
                level=self._extract_level(),
                format=self.get_value('log_format'),
                datefmt=self.get_value('log_date_format'))
            logging.info(
                "\n\n\nLogging started {}"
                .format(datetime.datetime.now().strftime("%I:%M:%S %p")))

def configure():
    LogConfig().configure()



#!/usr/bin/env python

import argparse
import os
import stat

from . import arg_helper
from ..lib import injection
from ..lib import settings
from ..parser.parse import Parser

from . import config_values

def program_code(instructions):
    output = ''
    with open(os.path.join(
            'bardolph', 'controller', 'lsc_template.py')) as srce:
        for line in srce:
            if line.find('#instructions') > -1:
                output += instructions
            else:
                output += line
    return output
    
def instruction_text(file_name):
    program = Parser().load(file_name)
    if program is None:
        return None
    return ',\n'.join(map(lambda inst: inst.as_list_text(), program))

def output_python(output_text, output_name=None):
    if output_name is None:
        print(output_text)
    else:
        out_file = open(output_name, 'w')
        out_file.write(output_text)
        out_file.close()
        os.chmod(output_name, stat.S_IRWXU | stat.S_IRGRP | stat.S_IROTH)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('file', help='name of the script file')
    arg_helper.add_o_argument(parser)
    args = parser.parse_args()

    injection.configure()
    settings.use_base(config_values.functional).configure()

    input_file = args.file
    print(input_file) 
    program = instruction_text(input_file)
    if program is not None:
        output_python(program_code(program), arg_helper.get_output_file(args))

if __name__ == '__main__':
    main()



#!/usr/bin/env python

import argparse
import logging

from ..controller.instruction import Instruction, OpCode, Operand
from ..controller.instruction import Register, SetOp
from ..controller.units import UnitMode, Units
from . import lex
from ..lib.time_pattern import TimePattern
from .token_types import TokenTypes


class Parser:
    def __init__(self):
        self._lexer = None
        self._error_output = ''
        self._light_state = {}
        self._name = None
        self._current_token_type = None
        self._current_token = None
        self._op_code = OpCode.NOP
        self._symbol_table = {}
        self._code = []
        self._unit_mode = UnitMode.LOGICAL

    def parse(self, input_string):
        self._code.clear()
        self._error_output = ''
        self._lexer = lex.Lex(input_string)
        self._next_token()
        success = self._script()
        self._lexer = None
        if not success:
            return None
        self._optimize()
        return self._code

    def load(self, file_name):
        logging.debug('File name: {}'.format(file_name))
        try:
            srce = open(file_name, 'r')
            input_string = srce.read()
            srce.close()
            return self.parse(input_string)
        except FileNotFoundError:
            logging.error('Error: file {} not found.'.format(file_name))
        except OSError:
            logging.error('Error accessing file {}'.format(file_name))

    def get_errors(self):
        return self._error_output

    def _script(self):
        return self._body() and self._eof()

    def _body(self):
        succeeded = True
        while succeeded and self._current_token_type != TokenTypes.EOF:
            succeeded = self._command()
        return succeeded

    def _eof(self):
        if self._current_token_type != TokenTypes.EOF:
            return self._trigger_error("Didn't get to end of file.")
        return True

    def _command(self):
        return {
            TokenTypes.DEFINE: self._definition,
            TokenTypes.GET: self._get,
            TokenTypes.OFF: self._power_off,
            TokenTypes.ON: self._power_on,
            TokenTypes.PAUSE: self._pause,
            TokenTypes.REGISTER: self._set_reg,
            TokenTypes.SET: self._set,
            TokenTypes.UNITS: self._set_units,
        }.get(self._current_token_type, self._syntax_error)()

    def _set_reg(self):
        self._name = self._current_token
        register = Register[self._name.upper()]
        if register == Register.TIME:
            return self._time()
        if register == Register.DURATION:
            return self._duration()
                
        self._next_token()
        if self._current_token_type == TokenTypes.NUMBER:
            try:
                value = round(float(self._current_token))
            except ValueError:
                return self._token_error('Invalid number: "{}"')
        elif self._current_token_type == TokenTypes.LITERAL:
            value = self._current_token
        elif self._current_token in self._symbol_table:
            value = self._symbol_table[self._current_token]
        else:
            return self._token_error('Unknown parameter value: "{}"')
        
        self._add_reg_instruction(register, value)
        return self._next_token()

    def _add_reg_instruction(self, reg, value):
        units = Units()
        if self._unit_mode == UnitMode.LOGICAL:
            value = units.as_raw(reg, value)
            if units.has_range(reg):
                (min_val, max_val) = units.get_range(reg)
                if value < min_val or value > max_val:
                    if self._unit_mode == UnitMode.LOGICAL:
                        min_val = units.as_logical(reg, min_val)
                        max_val = units.as_logical(reg, max_val)
                    return self._trigger_error(
                        '{} must be between {} and {}'.format(
                            reg.name.lower(), min_val, max_val))
        self._add_instruction(OpCode.SET_REG, reg, value)

    def _set_units(self):
        self._next_token()
        mode = {
            TokenTypes.RAW: UnitMode.RAW,
            TokenTypes.LOGICAL: UnitMode.LOGICAL
        }.get(self._current_token_type, None)

        if mode is None:
            return self._trigger_error(
                'Invalid parameter "{}" for units.'.format(self._current_token))

        self._unit_mode = mode
        return self._next_token()

    def _set(self):
        return self._action(OpCode.COLOR)

    def _get(self):
        self._next_token()
        if self._current_token_type == TokenTypes.LITERAL:
            self._name = self._current_token
        elif self._current_token in self._symbol_table:
            self._name = self._symbol_table[self._current_token]
        else:
            return self._token_error('Needed light or zone for "get", got {}.')
        
        self._next_token()
        if self._current_token_type == TokenTypes.ZONE:
            operand = Operand.MZ_LIGHT
            self._next_token()
            if self._current_token_type != TokenTypes.NUMBER:
                return self._token_error('Expected zone number, got "{}"')
            self._add_instruction(
                OpCode.SET_REG, Register.ZONES, (self._current_int(), None))
            self._next_token()
        else:
            operand = Operand.LIGHT

        self._add_instruction(OpCode.SET_REG, Register.NAME, self._name)
        self._add_instruction(OpCode.SET_REG, Register.OPERAND, operand)
        self._add_instruction(OpCode.GET_COLOR)
        return True

    def _power_on(self):
        self._add_instruction(OpCode.SET_REG, Register.POWER, True)
        return self._action(OpCode.POWER)

    def _power_off(self):
        self._add_instruction(OpCode.SET_REG, Register.POWER, False)
        return self._action(OpCode.POWER)

    def _pause(self):
        self._add_instruction(OpCode.PAUSE)
        self._next_token()
        return True
    
    def _time(self):
        self._next_token()
        if self._current_token_type == TokenTypes.AT:
            return self._process_time_patterns()

        if self._current_token_type == TokenTypes.NUMBER:
            value = self._normalized_time()
        elif self._current_token in self._symbol_table:
            value = self._normalized_time(
                self._symbol_table[self._current_token])
        else:
            return self._time_spec_error()
        self._add_reg_instruction(Register.TIME, value)
        return self._next_token()
        
    def _duration(self):
        self._next_token()
        if self._current_token_type == TokenTypes.NUMBER:
            value = self._normalized_time()
        elif self._current_token in self._symbol_table:
            value = self._normalized_time(
                self._symbol_table[self._current_token])
        else:
            return self._token_error("Expected number for duration, got {}.")
        self._add_reg_instruction(Register.DURATION, value)
        return self._next_token()
        
    def _process_time_patterns(self):
        time_pattern = self._next_time_pattern()
        if time_pattern is None:
            return self._time_spec_error()
        self._add_instruction(
            OpCode.TIME_PATTERN, SetOp.INIT, time_pattern)
        self._next_token()

        while self._current_token_type == TokenTypes.OR:
            time_pattern = self._next_time_pattern()
            if time_pattern is None:
                return self._time_spec_error()  
            self._add_instruction(
                OpCode.TIME_PATTERN, SetOp.UNION, time_pattern)
            self._next_token()

        return True;

    def _next_time_pattern(self):
        self._next_token()
        if self._current_token_type == TokenTypes.TIME_PATTERN:
            pattern_string = self._current_token
        else:
            pattern_string = self._symbol_table.get(self._current_token, None)
        if pattern_string is None:
            return None
        time_pattern = TimePattern.from_string(pattern_string)
        return time_pattern  
    
    def _action(self, op_code):
        """ op_code: COLOR, GET_COLOR, or POWER """
        self._op_code = op_code
        if self._op_code != OpCode.GET_COLOR:
            self._add_instruction(OpCode.TIME_WAIT)

        self._next_token()
        if self._current_token_type == TokenTypes.ALL:
            return self._all_operand()
        return self._operand_list()
    
    def _all_operand(self):
        self._add_instruction(OpCode.SET_REG, Register.NAME, None)
        self._add_instruction(OpCode.SET_REG, Register.OPERAND, Operand.ALL)
        self._add_instruction(self._op_code)
        return self._next_token()

    def _operand_list(self):
        if not self._operand():
            return self._syntax_error()

        self._add_instruction(self._op_code)
                
        while self._current_token_type == TokenTypes.AND:
            self._next_token()
            if not self._operand():
                return False
            self._add_instruction(self._op_code)
        return True
    
    def _operand(self):
        """ A group, location, or light with an optional set of zones. """
        if self._current_token_type == TokenTypes.GROUP:
            operand = Operand.GROUP
            self._next_token()
        elif self._current_token_type == TokenTypes.LOCATION:
            operand = Operand.LOCATION
            self._next_token()
        else:
            operand = Operand.LIGHT
            
        if self._current_token_type == TokenTypes.LITERAL:
            self._name = self._current_token
        elif self._current_token in self._symbol_table:
            self._name = self._symbol_table[self._current_token]
        else:
            return self._token_error('Unknown variable: {}')
        self._add_instruction(OpCode.SET_REG, Register.NAME, self._name)

        self._next_token()
        if self._current_token_type == TokenTypes.ZONE:
            if not self._zone_range():
                return False
            operand = Operand.MZ_LIGHT

        self._add_instruction(OpCode.SET_REG, Register.OPERAND, operand)
        return True

    def _zone_range(self):
        if self._op_code != OpCode.COLOR:
            return self._trigger_error('Zones not supported for {}'.format(
                self._op_code.name.tolower()))

        self._next_token()
        if self._current_token_type != TokenTypes.NUMBER:
            return self._token_error('Expected zone number, got "{}"')

        start_zone = self._current_int()
        self._next_token()

        if self._current_token_type == TokenTypes.NUMBER:
            end_zone = self._current_int()
            self._next_token()
        else:
            end_zone = start_zone + 1
            
        self._add_instruction(
            OpCode.SET_REG, Register.ZONES, (start_zone, end_zone))
        return True

    def _and(self):
        self._next_token()
        if not self._operand_name():
            return False
        self._add_instruction(OpCode.SET_REG, Register.NAME, self._name)
        self._add_instruction(self._op_code)
        return True

    def _definition(self):
        self._next_token()
        if self._current_token_type in [
                TokenTypes.LITERAL, TokenTypes.NUMBER, TokenTypes.TIME_PATTERN]:
            return self._token_error('Unexpected literal: {}')

        var_name = self._current_token
        self._next_token()
        if self._current_token_type == TokenTypes.NUMBER:
            value = int(self._current_token)
        elif self._current_token_type == TokenTypes.LITERAL:
            value = self._current_token
        elif self._current_token_type == TokenTypes.TIME_PATTERN:
            value = TimePattern.from_string(self._current_token)
            if value is None:
                return self._time_spec_error()
        elif self._current_token in self._symbol_table:
            value = self._symbol_table[self._current_token]
        else:
            return self._token_error('Unknown term: "{}"')

        self._symbol_table[var_name] = value
        self._next_token()
        return True

    def _add_instruction(self, op_code, param0=None, param1=None):
        self._code.append(Instruction(op_code, param0, param1))

    def _add_message(self, message):
        self._error_output += '{}\n'.format(message)

    def _trigger_error(self, message):
        full_message = 'Line {}: {}'.format(
            self._lexer.get_line_number(), message)
        logging.error(full_message)
        self._add_message(full_message)
        return False

    def _next_token(self):
        (self._current_token_type,
         self._current_token) = self._lexer.next_token()
        return True
    
    def _current_int(self):
        if self._current_token_type != TokenTypes.NUMBER:
            return None
        return round(float(self._current_token))

    def _token_error(self, message_format):
        return self._trigger_error(message_format.format(self._current_token))

    def _syntax_error(self):
        return self._token_error('Unexpected input "{}"')

    def _time_spec_error(self):
        return self._token_error('Invalid time specification: "{}"')  

    def _normalized_time(self, token=None):
        if token is None:
            token = self._current_token
        if self._unit_mode == UnitMode.LOGICAL:
            value = round(float(token) * 1000.0)
        else:
            value = int(token)            
        return value
    
    def _optimize(self):
        """
        Eliminate SET_REG if it has the same value as the previous SET_REG
        instruction.
        
        For this instruction, param0 is the name of the register, and param1 is
        the value to which the register is set.

        Any GET_COLOR instruction clears out the previous value cache.
        """
        opt = []
        prev_value = {}
        for inst in self._code:
            if inst.op_code == OpCode.GET_COLOR:
                prev_value = {}
                opt.append(inst)
            if inst.op_code != OpCode.SET_REG:
                opt.append(inst)
            else:
                key = str(inst.param0)
                if key not in prev_value or prev_value[key] != inst.param1:
                    opt.append(inst)
                    prev_value[key] = inst.param1
        self._code = opt


def main():
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument('file', help='name of the script file')
    args = arg_parser.parse_args()

    logging.basicConfig(
        level=logging.DEBUG,
        format='%(filename)s(%(lineno)d) %(funcName)s(): %(message)s')
    parser = Parser()
    output_code = parser.load(args.file)
    if output_code:
        for inst in output_code:
            print(inst)
    else:
        print(parser.get_errors())


if __name__ == '__main__':
    main()

"""
    <script> ::= <body> <EOF>
    <body> ::= <command> *
    <command> ::=
        "brightness" <set_reg>
        | "define" <definition>
        | "duration" <set_reg>
        | "hue" <set_reg>
        | "kelvin" <set_reg>
        | "off" <power_off>
        | "on" <power_on>
        | "_pause" <pause>
        | "saturation" <set_reg>
        | "_set" <set>
        | "units" <set_units>
        | "time" <time_spec>
    <set_reg> ::= <name> <number> | <name> <literal> | <name> <symbol>
    <set> ::= <action>
    <set_units> ::= "logical" | "raw"
    <get> ::= <action>
    <power_off> ::= <action>
    <power_on> ::= <action>
    <time_spec> ::= <number> | <time_pattern_set>
    <time_pattern_set> ::= <time_pattern> | <time_pattern> "or" <time_pattern>
    <time_pattern> ::= <hour_pattern> ":" <minute_pattern>
    <hour_pattern> ::= <digit> | <digit><digit> | "*" <digit> |
                        <digit> "*" | "*"
    <minute_pattern> ::= <digit><digit> | <digit> "*" | "*" <digit> | "*"
    <action> ::= <op_code> <operand_list> | <op_code> "all"
    <operand_list> ::= <operand> | <operand> <and> *
    <operand> ::= <name> | "group" <name> | "location" <name>
    <name> ::= <literal> | <token>
    <and> ::= "and" <operand_name>
    <definition> ::= <token> <number> | <token> <literal>
    <literal> ::= "\"" <token> "\""
"""



import logging

from ..lib.job_control import Job
from .machine import Machine
from ..parser.parse import Parser


class ScriptRunner(Job):
    def __init__(self):
        self._program = None
        self._parser = Parser()
        self._machine = Machine()

    @classmethod
    def from_file(cls, file_name):
        new_instance = ScriptRunner()
        new_instance.load_file(file_name)
        return new_instance

    @classmethod
    def from_string(cls, script):
        new_instance = ScriptRunner()
        new_instance.load_string(script)
        return new_instance

    def load_file(self, file_name):
        self._program = self._parser.load(file_name)
        if self._program is None:
            logging.error(self._parser.get_errors())
        return self._program

    def load_string(self, input_string):
        self._program = self._parser.parse(input_string)
        if self._program is None:
            logging.error(self._parser.get_errors())
        return self._program
    
    @property
    def program(self):
        return self._program

    def execute(self):
        if self._program is not None:
            self._machine.run(self._program)

    def request_stop(self):
        self._machine.stop()



import collections
import logging
import threading


class Job:
    def execute(self): pass
    def request_stop(self): pass


class Promise:
    def __init__(self, job, callback):
        self._job = job
        self._callback = callback

    def execute(self):
        threading.Thread(
            target=self.execute_and_call).start()
        return self

    def execute_and_call(self):
        self._job.execute()
        return self._callback()

    def request_stop(self):
        self._job.request_stop()


class JobControl:
    def __init__(self, repeat=False):
        self._promise = None
        self._queue = collections.deque()
        self._repeat = repeat
        self._lock = threading.RLock()

    def set_repeat(self, repeat):
        self._repeat = repeat

    def add_job(self, job):
        return self.enqueue_job(job, self._queue.append)

    def insert_job(self, job):
        return self.enqueue_job(job, self._queue.appendleft)

    def enqueue_job(self, job, append_fn):
        if not self._lock.acquire(True, 1.0):
            logging.error("Unable to acquire lock.")
        else:
            try:
                append_fn(job)
                if self._promise is None:
                    self.run_next_job()
            finally:
                self._lock.release()

    def has_jobs(self):
        return len(self._queue) > 0

    def run_next_job(self):
        if not self._lock.acquire(True, 1.0):
            logging.error("Unable to acquire lock.")
        else:
            try:
                if self._promise is None and len(self._queue) > 0:
                    next_job = self._queue.popleft()
                    if self._repeat:
                        self._queue.append(next_job)
                    self._promise = Promise(next_job, self.on_execution_done)
                    self._promise.execute()
            finally:
                self._lock.release()

    def on_execution_done(self):
        if not self._lock.acquire(True, 1.0):
            logging.error("Unable to acquire lock.")
        else:
            try:
                self._promise = None
            finally:
                self._lock.release()
            self.run_next_job()

    def request_stop(self):
        if not self._lock.acquire(True, 1.0):
            logging.error("Unable to acquire lock.")
        else:
            try:
                self._repeat = False
                self._queue.clear()
                promise = self._promise
                if promise is not None:
                    self._promise = None
                    promise.request_stop()
            finally:
                self._lock.release()

    def request_finish(self):
        """ Clear out the _queue but let the running job finish. """
        if not self._lock.acquire(True, 1.0):
            logging.error("Unable to acquire lock.")
        else:
            try:
                self._repeat = False
                self._queue.clear()
            finally:
                self._lock.release()



import re

from bardolph.lib.time_pattern import TimePattern
from .token_types import TokenTypes


class Lex:
    TOKEN_REGEX = re.compile(r'#.*$|".*?"|\S+')
    NUMBER_REGEX = re.compile(r'^[0-9]*\.?[0-9]+$')
    REG_REGEX = re.compile('hue|saturation|brightness|duration|time|kelvin')

    def __init__(self, input_string):
        self._lines = iter(input_string.split('\n'))
        self._line_num = 0
        self._tokens = None
        self._next_line()

    def _next_line(self):
        current_line = next(self._lines, None)
        if current_line is None:
            self._tokens = None
        else:
            self._line_num += 1
            self._tokens = self.TOKEN_REGEX.finditer(current_line)

    @classmethod
    def _unabbreviate(cls, token):
        return {
            'h': 'hue', 's': 'saturation', 'b': 'brightness', 'k': 'kelvin'
        }.get(token, token)

    def get_line_number(self):
        return self._line_num

    def next_token(self):
        token_type = None
        while token_type is None:
            match = None if self._tokens is None else next(self._tokens, None)
            while match is None:
                self._next_line()
                if self._tokens is None:
                    return (TokenTypes.EOF, '')
                match = next(self._tokens, None)

            token = Lex._unabbreviate(match.string[match.start():match.end()])
            if token[0] != '#':
                if token[0] == '"':
                    token = token[1:-1]
                    token_type = TokenTypes.LITERAL
                elif self.REG_REGEX.search(token):
                    token_type = TokenTypes.REGISTER
                elif TimePattern.REGEX.search(token):
                    token_type = TokenTypes.TIME_PATTERN
                elif self.NUMBER_REGEX.search(token):
                    token_type = TokenTypes.NUMBER
                else:
                    token_type = TokenTypes.__members__.get(
                        token.upper(), TokenTypes.UNKNOWN)

        return (token_type, token)
